#include "common/compat.h"

// generic components
#include "common/interconnect.h"
#include "components/echo/echo.h"
#include "components/arbiter/arbiter.h"
#include "components/map/map.h"
#include "components/reduce/reduce.h"
// needed for lm32 on altera
#include "components/ftdi/ftdi-byte-mode.h"
#include "components/byte-blaster/byte-blaster.h"
#include "components/jtag-bitbang/jtag-bitbang.h"
#include "components/jtag-chain/jtag-chain.h"
#include "components/altera-sld-hub/altera-sld-hub.h"
// LM32 contols
#include "components/lm32/lm32-break.h"
#include "components/lm32/lm32-reboot.h"
#include "components/lm32/lm32-read.h"
#include "components/lm32/lm32-write.h"
#include "components/lm32/lm32-drom.h"
// GDB 
#include "components/gdb-rsp/gdb-rsp.h"
#include "components/gdb-rsp/gdb-serial.h"

#include <iostream>
#include <iomanip>
#include <usb.h>
#include <stdio.h>
#include <signal.h>

void got(void* x, const Bits& b) {
  uint8_t byte = static_cast<uint8_t>(b);
  std::cout << (char)byte << std::flush;
}

LM32DROM* drom = 0;
void trapSIGINT(int sig) {
  drom->stop();
}

int main() {
  Interconnect ic;
  
  // Find usb blaster
  struct usb_bus* bus;
  struct usb_device* dev = 0;
  
  usb_init();
  if (usb_find_busses() < 0)
    return -1;
  if (usb_find_devices() < 0)
    return -1;
  
  for (bus = usb_get_busses(); bus; bus = bus->next) {
    for (dev = bus->devices; dev; dev = dev->next) { 
      printf("USB Device: %04x:%04x\n", dev->descriptor.idVendor, dev->descriptor.idProduct);
      if (dev->descriptor.idVendor  == 0x09fb && // Altera
          dev->descriptor.idProduct == 0x6001)   // USB-Blaster
        break;
/*
      struct usb_dev_handle* devx;
      char buf[256];
      if (!(devx = usb_open(dev))) { printf(" -> no permission\n"); continue; }
      if (usb_get_string_simple(devx, dev->descriptor.iProduct, buf, sizeof(buf)) > 0)
        printf(" -> %s\n", buf);
      if (usb_get_string_simple(devx, dev->descriptor.iSerialNumber, buf, sizeof(buf)) > 0)
        printf(" -> %s\n", buf);
      usb_close(devx);
*/
    }
    if (dev) break;
  }
  if (!dev) {
    std::cerr << "No USB-Blaster found\n";
    return -1;
  }
  
  Echo echo(&ic);
  Arbiter echos(&ic, true);
  FTDI_ByteMode ftdi(&ic, dev, 115200);
  Arbiter ftdi_w(&ic, false);
  ByteBlaster blaster_w(&ic, false);
  ByteBlaster blaster_rw(&ic, true);
  JTAG_BitBang banger(&ic);
  Reduce banger_ir(&ic);
  JTAG_Chain chain(&ic);

  try {
    ic.configure()
      .connect(echo.effect, ic.end().addEffect().port())
      .connect(echos.slave, echo.echo)
      .connect(ftdi.effect, ic.end().addEffect().port())
      .connect(ftdi_w.slave, ftdi.write)
      .connect(blaster_w.write_byte, ftdi_w.addClient().port())
      .connect(blaster_rw.read_byte, ftdi.read)
      .connect(blaster_rw.write_byte, ftdi_w.addClient().port())
      .connect(banger.JTAG_write, blaster_w.op)
      .connect(banger.JTAG_read_write, blaster_rw.op)
      .connect(banger_ir.slave_write, banger.IR_write)
      .connect(banger_ir.slave_read_write, banger.IR_read_write)
      .connect(chain.IR_write, banger_ir.master_write)
      .connect(chain.IR_read_write, banger_ir.master_read_write)
      .connect(chain.DR_write, banger.DR_write)
      .connect(chain.DR_read_write, banger.DR_read_write);
  } catch (const BadConnection& bc) {
    std::cerr << bc;
    return -1;
  }
  
  try {
    ic.reset();
    ic.execute();
  } catch (const ProtocolViolation& v) {
    std::cerr << v;
    return -1;
  }
  
  JTAG_Chain::IDCODEs codes = chain.detected();
  JTAG_Chain::IDCODEs::iterator i;
  
  std::cerr << "JTAG devices detected:" << std::endl;
  for (i = codes.begin(); i != codes.end(); ++i) {
    std::cerr << std::hex << *i << std::endl;
    if (*i == 0x20f30dd) break;
  }
  if (i == codes.end()) {
    std::cerr << "No Altera chip found!\n";
    return -1;
  }
  
  std::cerr << "Found Cyclone3!" << std::endl;
  
  Arbiter chain_ir(&ic, false);
  Map map_w(&ic, 10, false);
  Map map_rw(&ic, 10, true);
  Reduce map_ir(&ic);
  AlteraSldHub hub(&ic);
  
  chain_ir.slave.pinout() = Pinout("IR", 10);
  
  ic.configure()
    .connect(chain_ir.slave, *chain.findMaster("00000000/IR/write"))
    .connect(map_w.address, chain_ir.addClient().port())
    .connect(map_w.data, *chain.findMaster("00000000/DR/write"))
    .connect(map_rw.address, chain_ir.addClient().port())
    .connect(map_rw.data, *chain.findMaster("00000000/DR/read_write"))
    .connect(map_ir.slave_write, map_w.addRegister(14).port())
    .connect(map_ir.slave_read_write, map_rw.addRegister(14).port())
    .connect(hub.VIR_write, map_ir.master_write)
    .connect(hub.VIR_read_write, map_ir.master_read_write)
    .connect(hub.VDR_write, map_w.addRegister(12).port())
    .connect(hub.VDR_read_write, map_rw.addRegister(12).port());
    
  try {
    ic.reset();
    ic.execute();
  } catch (const ProtocolViolation& v) {
    std::cerr << v;
    return -1;
  }
  
  std::cerr << "Found LM32!" << std::endl;
  
  Arbiter dev0_ir(&ic, false);
  Map dev0_w(&ic, 1, false);
  Map dev0_rw(&ic, 1, true);
  Arbiter lm32_w(&ic, false);
  Arbiter lm32_rw(&ic, true);
  LM32Break lm32break(&ic);
  LM32Reboot lm32reboot(&ic);
  LM32Write lm32write(&ic, false);
  LM32Read lm32read(&ic, false);
  LM32DROM lm32drom(&ic);
  GDB_RSP gdb_rsp(&ic, 32, 32, lm32drom.PC, lm32drom.SP, lm32drom.SIG);
  GDB_Serial gdb_serial(&ic);
  
  SlavePort& uart = ic.head().addControl(uint8_pinout, (void*)0, &got).port();
  
  ic.configure()
    .connect(dev0_ir.slave, *hub.findMaster("00000000/IR/write"))
    .connect(dev0_w.address, dev0_ir.addClient().port())
    .connect(dev0_w.data, *hub.findMaster("00000000/DR/write"))
    .connect(dev0_rw.address, dev0_ir.addClient().port())
    .connect(dev0_rw.data, *hub.findMaster("00000000/DR/read_write"))
    .connect(lm32_w.slave, dev0_w.addRegister(0).port())
    .connect(lm32_rw.slave, dev0_rw.addRegister(0).port())
    .connect(lm32break.lm32_write, lm32_w.addClient().port())
    .connect(lm32break.lm32_read_write, lm32_rw.addClient().port())
    .connect(lm32reboot.lm32_write, lm32_w.addClient().port())
    .connect(lm32reboot.lm32_read_write, lm32_rw.addClient().port())
    .connect(lm32write.lm32_write, lm32_w.addClient().port())
    .connect(lm32write.lm32_read_write, lm32_rw.addClient().port())
    .connect(lm32read.lm32_write, lm32_w.addClient().port())
    .connect(lm32read.lm32_read_write, lm32_rw.addClient().port())
    .connect(lm32drom.echo, echos.addClient().port())
    .connect(lm32drom.uart_write, lm32write.write)
    .connect(lm32drom.uart_read, lm32read.read)
    .connect(lm32drom.uart_poll, lm32read.poll)
    .connect(lm32drom.jtag_break, lm32break.execute)
    .connect(gdb_rsp.echo, echos.addClient().port())
    .connect(gdb_rsp.register_address, lm32drom.register_address)
    .connect(gdb_rsp.register_write, lm32drom.register_write)
    .connect(gdb_rsp.register_read, lm32drom.register_read)
    .connect(gdb_rsp.memory_address, lm32drom.memory_address)
    .connect(gdb_rsp.memory_write, lm32drom.memory_write)
    .connect(gdb_rsp.memory_read, lm32drom.memory_read)
    .connect(gdb_rsp.breakpoint_type, lm32drom.breakpoint_type)
    .connect(gdb_rsp.breakpoint_add, lm32drom.breakpoint_add)
    .connect(gdb_rsp.breakpoint_del, lm32drom.breakpoint_del)
    .connect(gdb_rsp.run, lm32drom.run)
    .connect(gdb_serial.echo, echos.addClient().port())
    .connect(gdb_serial.packet, gdb_rsp.packet)
    .connect(uart, gdb_serial.uart);
  
  try {
    ic.reset();
    ic.execute();
  } catch (const ProtocolViolation& v) {
    std::cerr << v;
    return -1;
  }
  
  drom = &lm32drom;
  signal(SIGINT, &trapSIGINT);
  
  std::cerr << "Executing break -- and entering pipe-mode" << std::endl;
  
  uint8_t buf[1024];
  while (1) {
    // Poll device
    try {
      ic.execute();
    } catch (const ProtocolViolation& v) {
      std::cerr << v;
      return -1;
    }
    
    // Get next command
    int got = read(0, buf, sizeof(buf));
    if (got <= 0) break; // EOF
    for (int i = 0; i < got; ++i)
      uart(Bits(buf[i]));
  }
  
  return 0;
}
