#ifndef BITS_H
#define BITS_H

#include "integer.h"
#include "log2.h"

#include <cstring> // size_t, uintptr_t, ptrdiff_t
#include <iterator>
#include <iostream>

/* TODO: add allocator support */
class Bits {
  protected:
    typedef uintptr_t rep;
    
  public:
    struct Cell {
      public:
        operator bool () const;
        
        Cell& operator = (bool x);
        Cell& operator = (const Cell& x);
        void flip();
        
        bool operator == (const Cell& x) const;
        bool operator != (const Cell& x) const;
        bool operator <  (const Cell& x) const;
        bool operator >  (const Cell& x) const;
        bool operator <= (const Cell& x) const;
        bool operator >= (const Cell& x) const;
        
      protected:
        Cell(rep* ptr, rep mask);
        
        rep* ptr;
        rep mask;
      
      friend class Bits;
      friend class Iterator;
      friend class ConstIterator;
    };
    
    struct IteratorBase : public std::iterator<std::random_access_iterator_tag, bool> {
      public:
        bool operator == (const IteratorBase& i) const;
        bool operator != (const IteratorBase& i) const;
        bool operator <  (const IteratorBase& i) const;
        bool operator >  (const IteratorBase& i) const;
        bool operator <= (const IteratorBase& i) const;
        bool operator >= (const IteratorBase& i) const;
        
      protected:
        IteratorBase(rep* p, rep m);
        
        rep* ptr;
        rep mask;
        
        void inc();
        void dec();
        void add(difference_type i);
        
        typedef rep my_rep;
        static const log2<rep> rep_log2;
      
      friend difference_type operator - (const IteratorBase& x, const IteratorBase& y);
    };
    
    struct Iterator : public IteratorBase {
      public:
        typedef Cell		reference;
        typedef Cell*		pointer;
        typedef Iterator	iterator;
        
        Iterator();
        
        reference operator * () const;
        iterator& operator ++ ();
        iterator& operator -- ();
        iterator  operator ++ (int);
        iterator  operator -- (int);
        
        iterator& operator += (difference_type i);
        iterator& operator -= (difference_type i);
        iterator  operator +  (difference_type i) const;
        iterator  operator -  (difference_type i) const;
        
        reference operator [] (difference_type i) const;
      
      protected:
        Iterator(rep* ptr, rep mask);
      
      friend class Bits;
      friend Iterator operator + (difference_type i, const Iterator& x);
    };
    
    struct ConstIterator : public IteratorBase {
      public:
        typedef bool		reference;
        typedef bool		const_reference;
        typedef const bool*	pointer;
        typedef ConstIterator	const_iterator;
        
        ConstIterator();
        ConstIterator(const Iterator& i);
        
        const_reference operator * () const;
        const_iterator& operator ++ ();
        const_iterator& operator -- ();
        const_iterator  operator ++ (int);
        const_iterator  operator -- (int);
        
        const_iterator& operator += (difference_type i);
        const_iterator& operator -= (difference_type i);
        const_iterator  operator +  (difference_type i) const;
        const_iterator  operator -  (difference_type i) const;
        
        const_reference operator [] (difference_type i) const;
      
      protected:
        ConstIterator(rep* ptr, rep mask);
      
      friend class Bits;
      friend ConstIterator operator + (difference_type i, const ConstIterator& x);
    };
    
    typedef bool					value_type;
    typedef size_t					size_type;
    typedef Iterator::difference_type			diference_type;
    typedef Cell					reference;
    typedef bool					const_reference;
    typedef Cell*					pointer;
    typedef const bool*					const_pointer;
    typedef Iterator					iterator;
    typedef ConstIterator				const_iterator;
    typedef std::reverse_iterator<iterator>		reverse_iterator;
    typedef std::reverse_iterator<const_iterator>	const_reverse_iterator;
    
    Bits(const Bits& x);
    ~Bits();
    
    // Empty bits
    Bits();
    
    // Given length
    explicit Bits(size_type n, const bool& value);
    
    // Named constructors
    struct InvalidConversion { };
    explicit Bits(uint8_t x);
    explicit Bits(uint16_t x);
    explicit Bits(uint32_t x);
    explicit Bits(uint64_t x);
    explicit Bits(const std::string& x);
    
    // Reverse conversion
    operator uint8_t () const;
    operator uint16_t () const;
    operator uint32_t () const;
    operator uint64_t () const;
    operator std::string () const;
    
    uint8_t get_uint8(size_type i) const;
    void set_uint8(size_type i, uint8_t b);
    
    //template <typename InputIterator>
    //void assign(InputIterator first, InputIterator last);
    //void assign(size_type n, const bool& value);
    
    void swap(Bits& x);
    static void swap(Bits& x, Bits& y);
    Bits& operator = (const Bits& x);
    
    bool empty() const;
    size_type size() const;
    size_type capacity() const;
    void resize(size_type n);
    void clear();
    
    reference operator [] (size_type i);
    const_reference operator [] (size_type i) const;
    reference at(size_type i);
    const_reference at(size_type i) const;
    
    iterator begin();
    iterator end  ();
    const_iterator begin() const;
    const_iterator end  () const;
    
    reverse_iterator rbegin();
    reverse_iterator rend  ();
    const_reverse_iterator rbegin() const;
    const_reverse_iterator rend  () const;
    
    reference front();
    reference back ();
    const_reference front() const;
    const_reference back () const;
    
    bool operator == (const Bits& x) const;
    bool operator != (const Bits& x) const;
    bool operator <  (const Bits& x) const; // When one prefixs the other, the shorter is "smaller"
    bool operator >  (const Bits& x) const;
    bool operator <= (const Bits& x) const;
    bool operator >= (const Bits& x) const;
    
    /* Bitwise operations
     *   length comes from the left-side operand 
     *   the right hand-side is zero extended for the operation
     */
    Bits operator & (const Bits& x) const;
    Bits operator | (const Bits& x) const;
    Bits operator ^ (const Bits& x) const;
    Bits operator ~ () const;
    
    Bits& operator |= (const Bits& x);
    Bits& operator &= (const Bits& x);
    Bits& operator ^= (const Bits& x);
    
    // insert, erase       -- need linear time and are not supported
    // push_back, pop_back -- require reserved size and are not supported
    
  protected:
    static size_type rep_cells(size_type bits);
    rep* alloc(size_type cells);
    rep* ptr() const;
    static size_type shift_cap(size_type shift, size_type size);
    
    union Un {
      rep  local;
      rep* buffer;
    };
    
    size_type length;
    mutable Un un;
  
  friend std::ostream& operator << (std::ostream& o, const Bits& b);
};

/* Implementation is all inlined for speed */
#include "bits.tcc"

#endif
