#include "compat.h" // precedes all system includes

#include "component.h"
#include "interconnect.h"
#include "pinouts/pinouts.h"

#include <iomanip>
#include <typeinfo>
#include <assert.h>

Component::Component(Interconnect* ic) 
 : interconnect_(ic), id(-1), masters_(), slaves_(),
   reset(*addMasterPort("reset", command_pinout)) {
}

Component::~Component() {
  assert (configurable());
  
  Interconnect::Configuration *cfg = 0;
  if (enabled()) cfg = new Interconnect::Configuration(interconnect()->configure());
  
  for (Component::Slaves::iterator i = slaves_begin(); i != slaves_end(); ++i)
    if (i->destination())
      interconnect()->disconnect(*i, *i->destination());
  
  for (Component::Masters::iterator i = masters_begin(); i != masters_end(); ++i)
    if (i->source())
      interconnect()->disconnect(*i->source(), *i);
  
  if (cfg) delete cfg;
}

void Component::execute() {
}

bool Component::configurable() const {
  return !enabled() || interconnect()->configurable();
}

Component::Masters::iterator Component::masters_begin() { return masters_.begin(); }
Component::Masters::iterator Component::masters_end()   { return masters_.end();   }
Component::Slaves::iterator  Component::slaves_begin()  { return slaves_ .begin(); }
Component::Slaves::iterator  Component::slaves_end()    { return slaves_ .end();   }

Component::Masters::iterator Component::findMaster(const std::string& name) {
  for (Masters::iterator i = masters_begin(); i != masters_end(); ++i)
    if (i->name() == name)
      return i;
  return masters_end();
}

Component::Slaves::iterator Component::findSlave(const std::string& name) {
  for (Slaves::iterator i = slaves_begin(); i != slaves_end(); ++i)
    if (i->name() == name)
      return i;
  return slaves_end();
}

Component::Masters::iterator Component::addMasterPort(const std::string& name, const Pinout& pinout) {
  masters_.push_back(MasterPort(name, pinout, this));
  return --masters_end();
}

Component::Slaves::iterator Component::addSlavePort (const std::string& name, const Pinout& pinout, bool optional) {
  assert (optional || configurable());
  
  Interconnect::Configuration *cfg = 0;
  if (!optional && enabled()) cfg = new Interconnect::Configuration(interconnect()->configure());
  
  slaves_.push_back(SlavePort(name, pinout, this, optional));
  
  if (cfg) delete cfg;
  
  return --slaves_end();
}

void Component::removeMasterPort(Component::Masters::iterator m) {
  assert (configurable() || !m->source());
  
  Interconnect::Configuration *cfg = 0;
  if (m->source() && enabled()) cfg = new Interconnect::Configuration(interconnect()->configure());
  
  if (m->source()) interconnect()->disconnect(*m->source(), *m);
  masters_.erase(m);
  
  if (cfg) delete cfg;
}

void Component::removeSlavePort(Component::Slaves::iterator s) {
  assert (configurable() || !s->destination());
  
  Interconnect::Configuration *cfg = 0;
  if (s->destination() && enabled()) cfg = new Interconnect::Configuration(interconnect()->configure());
  
  if (s->destination()) interconnect()->disconnect(*s, *s->destination());
  slaves_.erase(s);
  
  if (cfg) delete cfg;
}

std::ostream& operator << (std::ostream& o, const Component& c) {
  o << "Component(" << typeid(c).name() << "):" << std::endl;
  for (Component::Masters::const_iterator i = c.masters_.begin(); i != c.masters_.end(); ++i)
    o << "   . " << *i << std::endl;;
  o << "  =>" << std::endl;
  for (Component::Slaves::const_iterator i = c.slaves_.begin(); i != c.slaves_.end(); ++i)
    o << "   . " << *i << std::endl;
  return o;
}
