#ifndef COMPONENT_H
#define COMPONENT_H

#include "recvqueue.h"
#include "sendqueue.h"
#include "slaveport.h"
#include "masterport.h"
#include "protocolviolation.h"

#include <string>
#include <list>
#include <iostream>

class Interconnect;

/* A component represents a protocol node.
 * 
 * It has masters, which send commands to it.
 * It has slaves which to which it sends the translated commands.
 * 
 * In turn, the slaves may receive data which they send back via
 * recv. The component then maps that data to the masters.
 *
 * A component may provide multiple master ports; for example, JTAG
 * provides an IR and DR port. Similarly, it may need multiple slave ports.
 * Derived components specify their ports via add{Master/Slave}Port.
 *
 * The situation is somewhat complicated by buffering.
 * Many components combine multiple commands into commands sent to slaves.
 * For example, a bit FIFO might need to combine 8 write commands into 1.
 * Thus, the send and recv methods operate on queues.
 * 
 * Components can safely read/write to slave devices while processing a 
 * queue without worrying about cross-component ordering. The scheduler
 * code manages the queues behind the scenes to ensure maximal queueing
 * without violating the execution order.
 */
class Component {
  public:
    typedef std::list<MasterPort> Masters;
    typedef std::list<SlavePort> Slaves;
    
  private:
    Interconnect* interconnect_;
    int id; // x=-1 := disabled, x>=0 := assigned position x in DAG
  
    Masters masters_;
    Slaves slaves_;
  
  protected:
    /* Add/remove ports
     *   non-optional slaves can only added while configurable()
     *   masters and optional slaves may be added at any time
     *   connected masters/slaves may only be removed while configurable()
     */
    Masters::iterator addMasterPort(const std::string& name, const Pinout& pinout);
    Slaves::iterator  addSlavePort (const std::string& name, const Pinout& pinout, bool optional = false);
    void removeMasterPort(Masters::iterator m);
    void removeSlavePort(Slaves::iterator s);
    
  public:
    /* Attach to an owning interconnect.
     * Components should not be allowed to outlive the interconnect.
     */
    Component(Interconnect* ic);
    
    /* Remove the DISABLED component from the interconnect.
     * Assertion failure if !configurable.
     */
    virtual ~Component();
    
    /* Is this component enabled in the interconnect
     * ie: Every non-optional slave is connected to an enabled() component.
     */
    bool enabled() const;
    
    /* Can this component be configured?
     * ie: !enabled() or interconnect()->configurable()
     */
    bool configurable() const;
    
    /* Process received bits read from slave devices.
     * The entire recv queue MUST be processed before returning.
     * If there is a problem, throw ProtocolViolation
     * Note: it is an error to call 'send' inside recv.
     */
    virtual void recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) = 0;
    
    /* Process bits sent from a master device.
     * Normally, the entire send queue SHOULD be processed.
     * However, if the results of a read are needed before part of the
     * queue can be further processed, a short read is permitted.
     * Returned is the position in the stalled command queue.
     * If there is a problem, throw ProtocolViolation
     * Note: it is an error to call 'recv' inside send.
     */
    virtual SendQueue::const_iterator send(SendQueue::const_iterator i, SendQueue::const_iterator end) = 0;
    
    /* Every component has a master which resets internal state
     * and restarts the controlled device. It is automatically
     * connected inside the interconnect.
     */
    MasterPort& reset;
    
    /* Find a master/slave based on its name */
    Masters::iterator findMaster(const std::string& name);
    Slaves::iterator  findSlave (const std::string& name);
    
    /* Walk over all the connections */
    Masters::iterator masters_begin();
    Masters::iterator masters_end();
    
    Slaves::iterator slaves_begin();
    Slaves::iterator slaves_end();
    
    /* Retrieve owner */
    Interconnect* interconnect();
    const Interconnect* interconnect() const;
    
    /* For components which cause external side-effects, implement execute.
     * Most components just logically translate their input queue to output.
     * However, some components write to the screen/USB/serial/etc, a side-effect.
     * Components with external-to-phase side-effects need special handling:
     *
     * 1) Declare a slave called 'endOfChain' to connect to Interconnect::end().addEffect().
     * 2) Whenever 'send' is invoked, queue up the work that would have been done.
     * 3) Later execute_send is invoked to perform the queued work.
     * 
     * execute is invoked once for each endOfChain.write (with empty Bits)
     *   ... so each queued side-effect should be matched with an endOfChain.write
     * 
     * recv is invoked once for each endOfChain.read, after all execute calls
     *   ... so report responses detected in .send to endOfChain.read
     *   ... queue any data collected during execute
     *   ... pop the queue on each invocation of .recv
     */
    virtual void execute(); // Defaults to doing nothing
  
  friend class Interconnect;
  friend class SlavePort;
  friend class MasterPort;
  
  friend std::ostream& operator << (std::ostream& o, const Component& c);
};

inline Interconnect* Component::interconnect() {
  return interconnect_;
}

inline const Interconnect* Component::interconnect() const {
  return interconnect_;
}

inline bool Component::enabled() const {
  return id >= 0;
}

#endif
