#include "compat.h" // precedes all system includes

#include "interconnect.h"

#include <queue>
#include <typeinfo>
#include <assert.h>

// #define DEBUG_TRANSFERS

Interconnect::Interconnect() 
 : locked(0), allowSend(false), allowRecv(false), end_(this), head_(this) {
}

Interconnect::~Interconnect() {
  ++locked;
}

bool Interconnect::configurable() const {
  return !allowSend && !allowRecv;
}

Interconnect::Configuration Interconnect::configure() {
  assert (configurable());
  return Configuration(this);
}

Interconnect::Configuration::Configuration(const Configuration& c) : ic(c.ic) {
  ++ic->locked;
}

Interconnect::Configuration::Configuration(Interconnect* ic_) : ic(ic_) { 
  if (!ic->locked) {
    /* Anything in the pipeline = death! */
    for (Interconnect::Nodes::iterator i = ic->nodes.begin(); i != ic->nodes.end(); ++i) {
      assert (i->in_recv.empty());
      assert (i->in_send.empty());
      assert (i->out_recv.empty());
      assert (i->out_send.empty());
    }
      
    /* Clear all nodes and disconnect reset */
    for (Interconnect::Nodes::iterator i = ic->nodes.begin(); i != ic->nodes.end(); ++i) {
      i->component->id = -1;
      if (i->component != &ic->head()) {
        assert (i->component->reset.source());
        disconnect(*i->component->reset.source(), i->component->reset);
      }
    }
    ic->nodes.clear();
    
    /* Remove all reset slaves */
    HeadOfChain& head = ic->head();
    Component::Slaves::iterator i, j;
    for (i = head.slaves_begin(); i != head.slaves_end(); i = j) {
      j = i;
      ++j;
      if (!i->optional())
        head.removeSlavePort(i);
    }
  }
  ++ic->locked;
}

Interconnect::Configuration::~Configuration() {
  if (--ic->locked == 0)
    ic->sort();
}

void Interconnect::Configuration::swap(Interconnect::Configuration& c) {
  std::swap(c.ic, ic);
}

Interconnect::Configuration& Interconnect::Configuration::operator = (const Interconnect::Configuration& c) {
  Configuration copy(c);
  swap (copy);
  return *this;
}

Interconnect::Configuration& Interconnect::Configuration::connect(SlavePort& s, MasterPort& m) {
  ic->connect(s, m);
  return *this;
}

void Interconnect::connect(SlavePort& s, MasterPort& m) {
  if (m.source())
    throw BadConnection(&m, &s, "Master already has a slave");
  if (s.destination())
    throw BadConnection(&m, &s, "Slave already has a master");
  
  if (m.owner()->interconnect() != this)
    throw BadConnection(&m, &s, "Master's component is in another interconnect");
  if (s.owner()->interconnect() != this)
    throw BadConnection(&m, &s, "Slave's component is in another interconnect");
  
  s.master_ = &m;
  m.slave_ = &s;
}

Interconnect::Configuration& Interconnect::Configuration::disconnect(SlavePort& s, MasterPort& m) {
  ic->disconnect(s, m);
  return *this;
}

void Interconnect::disconnect(SlavePort& s, MasterPort& m) {
  if (m.owner()->interconnect() != this)
    throw BadConnection(&m, &s, "Master's component is in another interconnect");
  if (s.owner()->interconnect() != this)
    throw BadConnection(&m, &s, "Slave's component is in another interconnect");
  if (s.destination() != &m)
    throw BadConnection(&m, &s, "Master and slave are not connected");
  
  s.master_ = 0;
  m.slave_ = 0;
}

struct SortState {
  int unlinked_slaves;
  int unscheduled_parents;
};
typedef std::map<Component*, SortState> Explored;
typedef std::queue<Component*> Queue;

void Interconnect::sort() {
  assert (nodes.empty());
  
  // Track state needed during DFS and topological sort
  Explored explored;
  Queue queue;
  
  // Plan: BFS to find components with fully connected slaves
  // Rationale: only components with their slaves attached can be initialized
  // Approach: Count down the number of slaves a component has during BFS. 
  //           If zero, include it in the interconnect and explore its parents.
  
  // Start tracing components from the end-of-chain
  queue.push(&end());
  explored[&end()].unlinked_slaves = 0;
  
  while (!queue.empty()) {
    Component* c = queue.front();
    queue.pop();
    
    for (Component::Masters::iterator i = c->masters_begin(); i != c->masters_end(); ++i) {
      if (!i->source()) continue; // ignore disconnected port
      Component* d = i->source()->owner();
      if (explored.find(d) == explored.end()) {
        // Count the number of required slaves
        int& slaves = explored[d].unlinked_slaves;
        slaves = 0;
        for (Component::Slaves::iterator j = d->slaves_begin(); j != d->slaves_end(); ++j)
          if (!j->optional())
            ++slaves;
        
        queue.push(d);
      }
      if (!i->source()->optional() && --explored[d].unlinked_slaves == 0)
        queue.push(d);
    }
  }
  
  // We now clear out any components whose slaves were not fully attached
  for (Explored::iterator j, i = explored.begin(); i != explored.end(); i = j) {
    j =  i;
    ++j;
    if (i->second.unlinked_slaves > 0)
      explored.erase(i);
    else if (i->first != &head())
      connect(head().addReset(), i->first->reset);
  }
  
  // Push a record for the head
  explored[&head()];
  
  // Now, determine how many masters each included component has
  // Exclude any masters which were rejected for inclusion in the interconnect
  for (Explored::iterator i = explored.begin(); i != explored.end(); ++i) {
    i->second.unscheduled_parents = 0;
    Component::Masters::iterator mend = i->first->masters_end();
    for (Component::Masters::iterator j = i->first->masters_begin(); j != mend; ++j)
      if (j->source() && explored.find(j->source()->owner()) != explored.end())
        ++i->second.unscheduled_parents;
  }
  
  // Finally: Topologically sort it, starting from roots
  queue.push(&head());
  
  while (!queue.empty()) {
    Component* c = queue.front();
    queue.pop();
    
    // std::cerr << *c;
    
    // Add it to the sort
    c->id = nodes.size();
    nodes.push_back(Node());
    nodes.back().component = c;
    
    for (Component::Slaves::iterator i = c->slaves_begin(); i != c->slaves_end(); ++i) {
      // skip optional slaves
      if (!i->destination()) continue;
      
      Component* d = i->destination()->owner();
      assert (explored.find(d) != explored.end());
      if (--explored[d].unscheduled_parents == 0)
        queue.push(d);
    }
  }
  
  assert (nodes.size() >= 2);
  assert (nodes.front().component == &head());
  assert (nodes.back().component  == &end());
}

void Interconnect::reset() {
  // Initialize the chain
  for (Nodes::iterator i = --nodes.end(); i != nodes.begin(); --i) {
#ifdef DEBUG_TRANSFERS
    std::cout << "Reset:\n" << *i->component;
#endif
    (*i->component->reset.source())();
  }
}

void Interconnect::execute() {
  assert (!locked);
  assert (!allowSend && !allowRecv); // not recursive
  
  // So long as we have pending commands, do work!
  while (!nodes.empty() && !nodes.front().out_send.empty()) {
    int position = 0;
    int previous_full_out_send = -1;
    int previous_full_in_recv = -1;
    int nodes_end = nodes.size();
    int next_full_in_send = nodes_end;
    int next_full_out_recv = nodes_end;
    
    /* This algorithm uses the following invarients:
     *                       /-- position 
     *  [                    |                     ]
     *   ^^^ empty input ^^^   ^^^ empty output ^^^
     * 
     * A node is empty if both input and output are empty.
     *
     * Any node with output is blocked from moving that output to
     * the appropriate slave by an intervening non-empty node.
     * 'previous_full_out_send' lists all these blocked nodes.
     *
     * 'next_full_in_send' lists all non-empty nodes after position.
     */
    
    // Repeat read and write cycle until all data fed back to root
    while (position != -1) {
    
      allowSend = true;
      // Push data all the way until the end of the chain
      while (position != nodes_end) {
        Node& n = nodes[position];
        
        // If it waits on read, don't process again
        if (n.out_send.empty()) {
          // Process the input queue by the component
#ifdef DEBUG_TRANSFERS
          std::cout << "SEND " << typeid(*n.component).name() << ":" << std::endl;
#endif
          SendQueue::const_iterator cend = n.component->send(n.in_send.begin(), n.in_send.end());
          // Remove the const from the returned iterator
          SendQueue::iterator end = n.in_send.begin();
          std::advance(end, std::distance<SendQueue::const_iterator>(end, cend));
          // Erase the processed range
          n.in_send.erase(n.in_send.begin(), end);
        } else {
#ifdef DEBUG_TRANSFERS
          std::cout << "RESEND " << typeid(*n.component).name() << ":\n";
#endif
        }
        
        // Transfer all that we can without violating ordering
        int next;
        while (!n.out_send.empty() &&
               (next = n.out_send.front().master->owner()->id) <= next_full_in_send) {
          Node& nextn = nodes[next];
          // Move the front of the output queue to the end of the input queue
#ifdef DEBUG_TRANSFERS
          std::cout << " XFR => " << typeid(*nextn.component).name() << ": " << n.out_send.front().bits << "\n";
#endif
          nextn.in_send.splice(nextn.in_send.end(), n.out_send, n.out_send.begin());
          if (next < next_full_in_send) {
            nextn.next_full_in_send = next_full_in_send;
            next_full_in_send = next;
          }
        }
        
        if (!n.out_send.empty()) {
#ifdef DEBUG_TRANSFERS
          std::cout << " ... more to send, but blocked by queue\n";
#endif
          n.previous_full_out_send = previous_full_out_send;
          previous_full_out_send = position;
        } else if (!n.in_send.empty()) {
#ifdef DEBUG_TRANSFERS
          std::cout << " ... blocked on read\n";
#endif
          n.previous_full_out_send = previous_full_out_send;
          previous_full_out_send = -1;
        }
        
        // Move either forward or backward
        if (previous_full_out_send == -1 || // no prior node
            !n.in_send.empty() ||           // waiting for read
            !n.out_send.empty()) {          // output queue blocked
#ifdef DEBUG_TRANSFERS
          std::cout << " => advancing to next with input\n";
#endif
          position = next_full_in_send;
          if (position != nodes_end)
            next_full_in_send = nodes[position].next_full_in_send;
        } else {
#ifdef DEBUG_TRANSFERS
          std::cout << " => popping stack\n";
#endif
          position = previous_full_out_send;
          previous_full_out_send = nodes[position].previous_full_out_send;
        }
      }
      allowSend = false;
      
      /* We pushed all the data to the end of the chain that we could.
       * Now we propogate received data back up the first chain.
       * If a component blocked, it might have interrupted an operation
       * in-progress. If that happened, it stopped all further commands
       * from earlier in the chain that cross the blocking point.
       *
       * To prevent feeding back a half-finished read, we thus stop the
       * propogation the moment the receive chain has drained completely
       * to the point of a blocked reader.
       */
      --position;
      allowRecv = true;
      while (position != -1) {
        Node& n = nodes[position];
        
        // Process the input queue by the component
        if (n.out_recv.empty()) {
#ifdef DEBUG_TRANSFERS
          std::cout << "RECV: " << typeid(*n.component).name() << ":" << std::endl;
#endif
          n.component->recv(n.in_recv.begin(), n.in_recv.end());
          n.in_recv.clear();
        } else {
#ifdef DEBUG_TRANSFERS
          std::cout << "RERECV: " << typeid(*n.component).name() << ":\n";
#endif
        }
        
        // Transfer all that we can without violating ordering
        int prev;
        while (!n.out_recv.empty() &&
               (prev = n.out_recv.front().slave->owner()->id) >= previous_full_in_recv) {
          Node& prevn = nodes[prev];
#ifdef DEBUG_TRANSFERS
          std::cout << "  XFR => " << typeid(*prevn.component).name() << ": " << n.out_recv.front().bits << "\n";
#endif
          // Move the front of the output queue to the end of the input queue
          prevn.in_recv.splice(prevn.in_recv.end(), n.out_recv, n.out_recv.begin());
          if (prev > previous_full_in_recv) {
            prevn.previous_full_in_recv = previous_full_in_recv;
            previous_full_in_recv = prev;
          }
        }
        
        if (!n.out_recv.empty()) {
#ifdef DEBUG_TRANSFERS
          std::cout << " ... more to recv, but blocked by queue\n";
#endif
          n.next_full_out_recv = next_full_out_recv;
          next_full_out_recv = position;
        }

        // Detect that we've hit a read barrier, and resume send.
        if (next_full_out_recv == nodes_end && !n.in_send.empty()) {
#ifdef DEBUG_TRANSFERS
          std::cout << " => stopping to process blocked operation\n";
#endif
          previous_full_out_send = n.previous_full_out_send;
          break;
        }
        
        if (next_full_out_recv == nodes_end || // no other node to send
            !n.out_recv.empty()) {             // output blocked by queue
#ifdef DEBUG_TRANSFERS
          std::cout << " => advancing to prev with input\n";
#endif
          position = previous_full_in_recv;
          if (position != -1)
            previous_full_in_recv = nodes[position].previous_full_in_recv;
        } else {
#ifdef DEBUG_TRANSFERS
          std::cout << " => popping stack\n";
#endif
          position = next_full_out_recv;
          next_full_out_recv = nodes[position].next_full_out_recv;
        }
      }
      allowRecv = false;
    }
  }
}
