#ifndef INTERCONNECT_H
#define INTERCONNECT_H

#include "component.h"
#include "badconnection.h"
#include "components/end-of-chain/end-of-chain.h"
#include "components/head-of-chain/head-of-chain.h"

#include <vector>

class Interconnect {
  public:
    Interconnect();
    ~Interconnect();
    
    /* End of the chain; connect side-effects here */
    HeadOfChain& head();
    EndOfChain& end();
  
    /* Begin making changes to the interconnect. 
     * So long as a Configuration object exists:
     *    Interconnect::execute fails
     *    Every Component is !enabled
     * 
     * configure() will assert fail if there is anything in the queue
     * components are sorted, reconnected, and enabled on ~Configure
     */
    class Configuration {
      public:
        /* The usual C++ object ops */
        Configuration(const Configuration& c);
        ~Configuration();
        Configuration& operator = (const Configuration& c);
        void swap(Configuration& c);
        
        Configuration& connect(SlavePort& s, MasterPort& m);
        Configuration& disconnect(SlavePort& s, MasterPort& m);
      
      private:
        Configuration(Interconnect* ic);
        Interconnect* ic;
      friend class Interconnect;
    };
    
    /* An interconnect may not be configured inside execute() 
     * Attempts to configure it will assert fail.
     */
    bool configurable() const;
    Configuration configure();
    
    /* Drain queues, executing all commands.  
     * Should never be called recursively.
     */
    void execute();
    
    /* Issue reset to entire device chain.
     */
    void reset();
    
  private:
    unsigned int locked; // Number of outstanding Configuration objects
    bool allowSend;
    bool allowRecv;
    
    EndOfChain end_;
    HeadOfChain head_;
    
    // Precondition: will not affect topology
    void connect(SlavePort& s, MasterPort& m);
    void disconnect(SlavePort& s, MasterPort& m);
    
    struct Node {
      Component*		component;
      RecvQueue			in_recv;
      SendQueue			in_send;
      RecvQueue			out_recv;
      SendQueue			out_send;
      
      int				previous_full_out_send;
      int				next_full_in_send;
      int				previous_full_in_recv;
      int				next_full_out_recv;
    };
    typedef std::vector<Node> Nodes;
    Nodes nodes;
    
    /* Solve the graph problem to connect components */
    void sort();
    
  friend class SlavePort;
  friend class MasterPort;
  friend class Component;
};

inline HeadOfChain& Interconnect::head() {
  return head_;
}

inline EndOfChain& Interconnect::end() {
  return end_;
}

#endif
