#include "common/compat.h"

#include "altera-sld-hub.h"
#include "pinouts/pinouts.h"
#include "common/to_string.h"

#include <assert.h>

// #define DEBUG_HUB 1

Bits AlteraSldHub::HUB_INFO() {
  // 128 bits ought to cover it
  return Bits(128, false);
}

Bits AlteraSldHub::VIR_CAPTURE(int dev) {
  // Format: (m-3)x0 + addr + 011
  Bits command(n+m, false);
  // addr part: (0=hub, so dev+1)
  Bits::iterator o = command.begin()+3;
  for (unsigned int shift = dev+1; shift; shift >>= 1)
    *o++ = (shift&1)?1:0;
  // 011 part:
  command[1] = true;
  command[0] = true;
  
  return command;
}

Bits AlteraSldHub::VIR_WRITE(int dev, const Bits& b) {
  Bits command(n+m, false);
  Bits::iterator o = command.begin();
  
  // Fill in the IR 
  Bits::const_iterator ebits = b.end();
  for (Bits::const_iterator j = b.begin(); j != ebits; ++j)
    *o++ = *j;
  
  // Fill in the address
  for (unsigned int shift = dev+1; shift; shift >>= 1)
    *o++ = (shift&1)?1:0;
  
  return command;
}

Bits AlteraSldHub::VIR_READ(const Bits& b) {
  Bits trimmed(b);
  trimmed.resize(m);
  return trimmed;
}        

AlteraSldHub::AlteraSldHub(Interconnect* ic)
 : Bus(ic),
   VDR_write     (*addSlavePort("VDR/write",      shift_pinout)),
   VDR_read_write(*addSlavePort("VDR/read_write", shift_pinout)),
   VIR_write     (*addSlavePort("VIR/write",      shift_pinout)),
   VIR_read_write(*addSlavePort("VIR/read_write", shift_pinout)),
   m(0), n(0), nodes(0) {
}

void AlteraSldHub::recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) {
  while (i != end) {
    if (queue.empty())
      throw ProtocolViolation(this, "recv past queue end");
      
    MasterPort* master = queue.front();
    queue.pop();
    
    if (master == 0) {
      scanR(i);
      continue;
    }
    
    const std::string& name = master->name();
    bool ir = name[prefix] == 'I';
    assert (name.size() > (prefix+8));
    
    if (ir) {
      if (i->slave != &VIR_read_write)
        throw ProtocolViolation(this, "Expecting VIR recv");
      
#ifdef DEBUG_HUB
      std::cout << "Received VIR result " << i->bits << std::endl;
#endif
      (*master)(VIR_READ(i->bits));
    } else {
      if (i->slave != &VDR_read_write)
        throw ProtocolViolation(this, "Expecting VDR recv");
      
#ifdef DEBUG_HUB
      std::cout << "Received VDR result " << i->bits << std::endl;
#endif
      (*master)(i->bits);
    }
    ++i;
  }
}

SendQueue::const_iterator AlteraSldHub::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (i->master == &reset) {
      if (n == 0) {
        scanW();
        return i; // block
      }
      // already initialized
      continue;
    }
    
    assert (n > 0); // else could never have been called!
    
    const std::string& name = i->master->name();
    uint32_t dev = device(name);
    bool ir = name[prefix] == 'I';
    bool read = name.size() > (prefix+8);
    
    if (read) queue.push(i->master);
    
    if (ir) {
      if (i->bits.size() > m)
        throw ProtocolViolation(this, "IR write exceeds VIR supported size");
      
      // pad
      Bits bits(i->bits);
      bits.resize(m);
      
      if (read) {
        VIR_write(VIR_CAPTURE(dev));
        VIR_read_write(VIR_WRITE(dev, bits));
      } else { // write
        VIR_write(VIR_WRITE(dev, bits));
      }
    } else { // dr 
      if (read)
        VDR_read_write(i->bits);
      else
        VDR_write(i->bits);
    }
  }
  
  return i;
}

void AlteraSldHub::scanW() {
#ifdef DEBUG_HUB
  std::cout << "Scanning hub with HUB_INFO" << std::endl;
#endif
  
  // Put device into HUB_INFO mode
  VIR_write(HUB_INFO());
  
  // Read the 8x 4-bit nibble config
  Bits nib(4, false);
  for (int i = 0; i < 8; ++i) {
    VDR_read_write(nib);
  }
  
  // Expect an init response
  queue.push(0);
}

static void cr_load(RecvQueue::const_iterator& i, Bits& bits, int offset) {
  for (int j = 0; j < 4; ++j)
    bits[offset*4+j] = i->bits[j];
  ++i;
}

static log2<uint8_t> log8;
void AlteraSldHub::scanR(RecvQueue::const_iterator& i) {
  Bits controlRegister(32, false);
  
  for (int j = 0; j < 8; ++j)
    cr_load(i, controlRegister, j);
  
#ifdef DEBUG_HUB
      std::cout << "Received hub scan " << controlRegister << std::endl;
#endif
  
  uint32_t value = controlRegister;
  nodes   = (value >> 19) & 0xff;  /* nodes = bits 26..19 */
  m       = (value >>  0) & 0xff;  /* m = bits 7..0  */
  
  /* NOTE: The Altera documenation is WRONG.
   * It claims the field 7..0 holds (m+n).
   * However, I have verified with several IR widths that this is false.
   */
  
  int mf  = (value >>  8) & 0x7ff; /* manufacturer 18..8 */
  int ver = (value >> 27) & 0x1f;  /* version 31..27 */
  
  if (mf != 0x6e)
    throw ProtocolViolation(this, "Manufacturer of sld_hub is not Altera (" + to_string(mf) + ")");
  if (ver != 1)
    throw ProtocolViolation(this, "Unsupported hub version (" + to_string(ver) + ")");
  
  /* n = ceil(log2(nodes+1)) */
  unsigned int l = nodes;
  l |= (l >> 1);
  l |= (l >> 2);
  l |= (l >> 4);
  ++l;
  n = log8(l);
  
  /* I have confirmed that Quartus expands the sld_hub's IR to length n+3.
   * This is necessary since it must fit (address) + "011" for a VIR_CAPTURE.
   */
  if (m < n + 3)
    throw ProtocolViolation(this, "m < n+3 => impossible to use VIR_CAPTURE (" + to_string(m) + "," + to_string(n) + ")");
  
#ifdef DEBUG_HUB
      std::cout << "m=" << m << ", n=" << n << ", nodes=" << nodes << "\n";
#endif
  for (unsigned int j = 0; j < nodes; ++j) {
    addMasterPort(makeName(j, "IR/write"),      shift_pinout);
    addMasterPort(makeName(j, "IR/read_write"), shift_pinout);
    addMasterPort(makeName(j, "DR/write"),      shift_pinout);
    addMasterPort(makeName(j, "DR/read_write"), shift_pinout);
  }
}
