#include "common/compat.h"

#include "byte-blaster.h"
#include "pinouts/pinouts.h"

/* Byte-blaster protocol: */
#define BYTE_SHIFT_MODE	7
#define READ_BIT	6
#define LED_ACTIVE	5
#define TDI_BIT		4
#define nCS_BIT		3
#define nCE_BIT		2
#define TMS_BIT		1
#define TCK_BIT		0

/* Input: */
#define TDO_BIT		0

ByteBlaster::ByteBlaster(Interconnect* ic, bool rf)
 : Component(ic), 
   op(*addMasterPort("op", JTAG_pinout)),
   read_byte (*addSlavePort("read",  uint8_pinout, !rf)),
   write_byte(*addSlavePort("write", uint8_pinout)),
   TMS(JTAG_pinout.getPin("TMS")),
   TDI(JTAG_pinout.getPin("TDI")),
   TDO(JTAG_pinout.getPin("TDO")),
   read_flag(rf) {
}

void ByteBlaster::recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (i->slave == &write_byte) 
      throw ProtocolViolation(this, "Write port recv'd data!");
    
    if (queue.empty())
      throw ProtocolViolation(this, "Exhausted queue during recv");
    bool shift = queue.front();
    queue.pop();
    
    Bits out(JTAG_pinout.pins(), false);
    Bits::reference otdo = out[TDO];
    
    if (shift) {
      uint8_t byte = static_cast<uint8_t>(i->bits);
      for (int k = 0; k < 8; ++k) {
        otdo = (byte & 1) != 0;
        byte >>= 1;
        op(out);
      }
    } else {
      // Single read
      otdo = i->bits[TDO_BIT];
      op(out);
    }
  }
}

SendQueue::const_iterator ByteBlaster::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  while (i != end) {
    if (i->master == &reset) { ++i; continue; }
    
    // Look for the longest sequence of TMS=0
    SendQueue::const_iterator j;
    SendQueue::size_type len = 0;
    for (j = i; j != end; ++j, ++len)
      if (j->bits[TMS] || len >= 8*64)
        break;
    
    if (len >= 8) {
      // Byte-shift mode!
      SendQueue::size_type k, m, work = len / 8;
      if (work >= 64) work = 63;
      
      // For whatever stupidity, byte-shift mode does clock high-low (not low-high)
      // Thus, setup the nCS bit and first TDI safely
      Bits prep(uint8_pinout.pins(), false);
      prep[nCS_BIT] = true;
      prep[LED_ACTIVE] = true;
      prep[TDI_BIT] = i->bits[TDI];
      write_byte(prep);
      
      uint8_t count = work;
      Bits bsm(count);
      bsm[BYTE_SHIFT_MODE] = true;
      bsm[READ_BIT] = read_flag;
      
      write_byte(bsm);
      for (k = 0; k < work; ++k) {
        uint8_t byte = 0;
        for (m = 0; m < 8; ++m) {
          byte >>= 1;
          byte |= i->bits[TDI]?0x80:0;
          ++i;
        }
        write_byte(Bits(byte));
        
        if (read_byte.connected()) {
          queue.push(true);
          read_byte();
        }
      }
    } else {
      // More or less just a pin remapping for bitbang mode
      Bits out(uint8_pinout.pins(), false);
      const Bits& in = i->bits;

      Bits::reference otck = out[TCK_BIT];
      Bits::reference otms = out[TMS_BIT];
      Bits::reference otdi = out[TDI_BIT];
      Bits::reference orf  = out[READ_BIT];
      Bits::reference oled = out[LED_ACTIVE];
      
      bool tms = in[TMS];
      bool tdi = in[TDI];
      
      // JTAG samples on positive edge Move all the assignments to the end --- avoids aliasing
      otms = tms;
      otdi = tdi;
      
      otck = false;
      orf = read_flag;
      oled = true;
      write_byte(out); // write the pins
      
      otck = true;
      orf = false;
      oled = ++i != end;
      write_byte(out);
      
      // The result of this is what we wait for
      if (read_byte.connected()) {
        queue.push(false);
        read_byte();
      }
    }
  }
  return i;
}
