#include "common/compat.h"

#include "ftdi-byte-mode.h"
#include "pinouts/pinouts.h"

#include <assert.h>
#include <stdio.h>

#define MAX_RECV_BUF	64
#define MAX_SEND_BUF	64

// #define DEBUG_FTDI

FTDI_ByteMode::FTDI_ByteMode(Interconnect* ic, struct usb_device* dev, int rate)
 : Component(ic),
   write (*addMasterPort("write", uint8_pinout)),
   read  (*addMasterPort("read",  uint8_pinout)),
   effect(*addSlavePort("effect", command_pinout)),
   baudrate(rate)
{
  if (ftdi_init(&ftdi) != 0)
    throw ProtocolViolation(0, "Not enough memory to init ftdi");
  if (ftdi_usb_open_dev(&ftdi, dev) != 0)
    throw ProtocolViolation(this, std::string("init: ftdi_usb_open_dev: ") + ftdi_get_error_string(&ftdi));
}

FTDI_ByteMode::~FTDI_ByteMode() {
  ftdi_usb_close(&ftdi);
  ftdi_deinit(&ftdi);
}

void FTDI_ByteMode::recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (toRecv.empty())
      throw ProtocolViolation(this, "too many effects");
    
    const Operation& o = toRecv.front();
    
    if (o.buf) {
      for (unsigned int i = 0; i < o.len; ++i)
        read(Bits(o.buf[i]));
      delete [] o.buf;
    }
    
    toRecv.pop();
  }
}

void FTDI_ByteMode::flush_write(Operation& o) {
  toExecute.push(o);
  effect();
  o.buf = new unsigned char[MAX_SEND_BUF];
  o.len = 0;
}

void FTDI_ByteMode::flush_read(unsigned int& len) {
  Operation o;
  o.len = len;
  o.buf = 0;
  toExecute.push(o);
  effect(); 
  len = 0;
}

SendQueue::const_iterator FTDI_ByteMode::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  unsigned int in_len = 0;
  Operation o;
  o.len = 0;
  o.buf = new unsigned char[MAX_SEND_BUF];
  
  for (; i != end; ++i) {
    if (i->master == &reset) {
      if (ftdi_set_line_property2(&ftdi, BITS_8, STOP_BIT_1, NONE, BREAK_OFF) != 0) // 8N1
        throw ProtocolViolation(this, std::string("init: ftdi_set_line_property2: ") + ftdi_get_error_string(&ftdi));
      if (ftdi_disable_bitbang(&ftdi) != 0)  // byte mode
        throw ProtocolViolation(this, std::string("init: ftdi_disable_bitbang: ") + ftdi_get_error_string(&ftdi));
      if (ftdi_set_baudrate(&ftdi, baudrate) != 0)
        throw ProtocolViolation(this, std::string("init: ftdi_set_baudrate: ") + ftdi_get_error_string(&ftdi));
      continue;
    } else if (i->master == &read) {
      // Flush output before trying to read (else might block)
      if (++in_len == MAX_RECV_BUF) {
        if (o.len > 0) flush_write(o);
        flush_read(in_len);
      }
    } else {
      o.buf[o.len] = static_cast<unsigned char>(i->bits);
      if (++o.len == MAX_SEND_BUF) flush_write(o);
    }
  }
  
  if (o.len > 0) flush_write(o);
  if (in_len > 0) flush_read(in_len);
  
  delete [] o.buf;
  
  return i;
}

void FTDI_ByteMode::execute() {
  if (toExecute.empty())
    throw ProtocolViolation(this, "No pending side-effects in execute");
  Operation& o = toExecute.front();
  
  if (o.buf) {
    unsigned int sent = 0;
    int got = 0;
    while (sent != o.len && (got = ftdi_write_data(&ftdi, o.buf+sent, o.len-sent)) >= 0) {
#ifdef DEBUG_FTDI
      fprintf(stderr, "ftdi_write_data: %d, %d, %d:\n", o.len, sent, got);
#endif
      sent += got;
    }
    
#ifdef DEBUG_FTDI
    fprintf(stderr, "payload: ");
    for (unsigned int i = 0; i < sent; ++i)
      fprintf(stderr, "%02X ", o.buf[i]);
    fprintf(stderr, "\n");
#endif
    
    if (got < 0)
      throw ProtocolViolation(this, std::string("send: ftdi_write_data: ") + ftdi_get_error_string(&ftdi));
    
    delete [] o.buf;
    
    o.buf = 0;
    toRecv.push(o);
  } else {
    unsigned int recv = 0;
    int got = 0;
    o.buf = new unsigned char[o.len];
    
    while (recv != o.len && (got = ftdi_read_data(&ftdi, o.buf+recv, o.len-recv)) >= 0) {
#ifdef DEBUG_FTDI
      fprintf(stderr, "ftdi_read_data: %d, %d, %d:\n", o.len, recv, got);
#endif
      recv += got;
    }
    o.len = recv;
  
#ifdef DEBUG_FTDI
    fprintf(stderr, "payload: ");
    for (unsigned int i = 0; i < recv; ++i)
      fprintf(stderr, "%02X ", o.buf[i]);
    fprintf(stderr, "\n");
#endif
    
    toRecv.push(o);
    if (got < 0)
      throw ProtocolViolation(this, std::string("recv: ftdi_read_data: ") + ftdi_get_error_string(&ftdi));
  }
  
  toExecute.pop();
}

