#include "common/compat.h"

#include "gdb-rsp.h"
#include "pinouts/pinouts.h"

GDB_RSP::GDB_RSP(Interconnect* ic, int width, int r, int pc, int sp, int sig) :
  Component(ic),
  packet(*addMasterPort("uart", shift_pinout)),
  echo(*addSlavePort("echo", shift_pinout)),
  register_address(*addSlavePort("register/address", Pinout("address", 8))),
  register_write  (*addSlavePort("register/write",   Pinout("data",    width))),
  register_read   (*addSlavePort("register/read",    Pinout("data",    width))),
  memory_address  (*addSlavePort("memory/address",   Pinout("address", width))),
  memory_write    (*addSlavePort("memory/write",     Pinout("data",    8))),
  memory_read     (*addSlavePort("memory/read",      Pinout("data",    8))),
  breakpoint_type (*addSlavePort("breakpoint/type",  uint8_pinout)),
  breakpoint_add  (*addSlavePort("breakpoint/add",   Pinout("address", width))),
  breakpoint_del  (*addSlavePort("breakpoint/del",   Pinout("address", width))),
  run             (*addSlavePort("run",              command_pinout)),
  registers(r), PC(pc), SP(sp), SIG(sig) {
}

void GDB_RSP::recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (i->slave == &echo) {
      packet(i->bits);
    } else if (i->slave == &breakpoint_add || i->slave == &breakpoint_del) {
      packet(Bits(i->bits[0]?"OK":"E28"));
    } else {
      // memory or register read => both report hex
      if (queue.empty())
        throw ProtocolViolation(this, "queue underrun in recv");
      int len = queue.front();
      queue.pop();
      
      if (len == -1) {
        // status report
        if (i->slave != &register_read) throw ProtocolViolation(this, "expecting status SIG");
        const Bits& sig = (i++)->bits;
        if (i->slave != &register_read) throw ProtocolViolation(this, "expecting status PC");
        const Bits& pc = (i++)->bits;
        if (i->slave != &register_read) throw ProtocolViolation(this, "expecting status SP");
        const Bits& sp = i->bits;
        
        Bits sigb = sig;
        sigb.resize(8);
        
        std::string output;
        output.push_back('T');
        toHex(output, sigb);
        
        toHex(output, Bits(PC));
        output.push_back(':');
        toHex(output, pc);
        output.push_back(';');
        
        toHex(output, Bits(SP));
        output.push_back(':');
        toHex(output, sp);
        output.push_back(';');
        
        packet(Bits(output));
      } else {
        std::string output;
        output.reserve(len*(i->slave==&register_read?8:2));
        for (int j = 0; j < len; ++j, ++i)
          toHex(output, i->bits);
        packet(Bits(output));
        --i;
      }
    }
  }
}

SendQueue::const_iterator GDB_RSP::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (i->master == &reset) continue;
    runCommand(static_cast<std::string>(i->bits));
  }
  
  return i;
}

void GDB_RSP::runCommand(const std::string& command) {
  if (command.empty()) return;
  
  switch (command[0]) {
  case '?':
    cmdStatus(command);
    break;
  case 'g':
    cmdRegGetAll(command);
    break;
  case 'G':
    cmdRegSetAll(command);
    break;
  case 'm':
    cmdMemRead(command);
    break;
  case 'M':
    cmdMemWrite(command);
    break;
  case 'p':
    cmdRegGet(command);
    break;
  case 'P':
    cmdRegSet(command);
    break;
  case 'D':
  case 'k':
    cmdDetach(command);
    break;
  case 'c':
  case 's':
    cmdContinue(command);
    break;
  case 'z':
  case 'Z':
    cmdBreak(command);
    break;
  case 'r':
  case 'R':
    cmdReset(command);
    break;
  case 'q':
    cmdQuery(command);
    break;
  default:
    echo(Bits()); // empty response for no support
    break;
  }
}

void GDB_RSP::cmdStatus(const std::string& command) {
  register_address(Bits(SIG));
  register_read();
  register_address(Bits(PC));
  register_read();
  register_address(Bits(SP));
  register_read();
  queue.push(-1);
}

void GDB_RSP::cmdQuery(const std::string& command) {
  if (command == "qSupported") {
    echo(Bits("PacketSize=4000"));
  } else {
    echo(Bits());
  }
}

void GDB_RSP::cmdRegGetAll(const std::string& str) {
  for (uint8_t reg = 0; reg < registers; ++reg) {
    register_address(Bits(reg));
    register_read();
  }
  queue.push(registers);
}

void GDB_RSP::cmdRegSetAll(const std::string& str) {
  const char* ptr = str.data() + 1;
  
  uint8_t reg;
  for (reg = 0; reg < registers; ++reg) {
    Bits reg(register_write.pinout().pins(), false);
    if (fromHex(ptr, reg) != 0) break;
    
    register_address(Bits(reg));
    register_write(reg);
  }
  
  if (reg == registers) {
    echo(Bits("OK"));
  } else {
    echo(Bits("E22")); // EINVAL
  }
}

void GDB_RSP::cmdMemRead(const std::string& str) {
  const char* ptr = str.c_str() + 1;
  uint64_t addr, len;
  
  if (fromHex(ptr, addr) != 0 ||
      *ptr++ != ',' ||
      fromHex(ptr, len) != 0 ||
      *ptr != 0) {
    echo(Bits("E22")); // EINVAL
    return;
  }
  
  Bits a(addr);
  a.resize(memory_address.pinout().pins());
  memory_address(a);
  queue.push(len);
  
  for (uint64_t i = addr; i < addr+len; ++i)
    memory_read(); // !!! deal with faults?
}

void GDB_RSP::cmdMemWrite(const std::string& str) {
  const char* ptr = str.c_str() + 1;
  uint64_t addr, len;
  
  if (fromHex(ptr, addr) != 0 ||
      *ptr++ != ',' ||
      fromHex(ptr, len) != 0 ||
      *ptr++ != ':') {
    echo(Bits("E22")); // EINVAL
    return;
  }
  
  Bits a(addr);
  a.resize(memory_address.pinout().pins());
  memory_address(a);
  
  uint64_t i;
  for (i = addr; i < addr+len; ++i) {
    Bits b(8, false);
    if (fromHex(ptr, b) != 0) break;
    memory_write(b);
  }
  
  if (i != addr+len) {
    echo(Bits("E22")); // EINVAL
  } else {
    echo(Bits("OK"));
  }
}

void GDB_RSP::cmdRegGet(const std::string& str) {
  const char* ptr = str.c_str() + 1;
  uint64_t reg;
  
  if (fromHex(ptr, reg) != 0 ||
      *ptr != 0 ||
      reg > registers) {
    echo(Bits("E22"));
  } else {
    register_address(Bits(static_cast<uint8_t>(reg)));
    register_read();
    queue.push(1);
  }
}

void GDB_RSP::cmdRegSet(const std::string& str) {
  const char* ptr = str.c_str() + 1;
  uint64_t regn;
  Bits regv(register_write.pinout().pins(), false);
  
  if (fromHex(ptr, regn) != 0 ||
      *ptr++ != '=' ||
      fromHex(ptr, regv) != 0 ||
      *ptr != 0 ||
      regn > registers) {
    echo(Bits("E22"));
  } else {
    register_address(Bits(static_cast<uint8_t>(regn)));
    register_write(regv);
    echo(Bits("OK"));
  }
}

void GDB_RSP::cmdBreak(const std::string& str) {
  const char* ptr = str.c_str()+2;
  uint64_t addr, kind;
  
  if (*ptr++ != ',' || 
      fromHex(ptr, addr) != 0 ||
      *ptr++ != ',' ||
      fromHex(ptr, kind) != 0 ||
      *ptr != 0) {
    echo(Bits("E22"));
    return;
  }
  
  Bits x(addr);
  x.resize(breakpoint_add.pinout().pins());
  
  switch (str[1]) {
    case '1': { /* breakpoint */
      breakpoint_type(Bits(static_cast<uint8_t>(0)));
      (str[0]=='Z'?breakpoint_add:breakpoint_del)(x);
      echo(Bits("OK"));
      break;
    }
    case '2': { /* write watchpoint */
      breakpoint_type(Bits(static_cast<uint8_t>(2)));
      (str[0]=='Z'?breakpoint_add:breakpoint_del)(x);
      echo(Bits("OK"));
      break;
    }
    case '3': { /* read watchpoint */
      breakpoint_type(Bits(static_cast<uint8_t>(1)));
      (str[0]=='Z'?breakpoint_add:breakpoint_del)(x);
      echo(Bits("OK"));
      break;
    }
    case '4': { /* access watchpoint */
      breakpoint_type(Bits(static_cast<uint8_t>(3)));
      (str[0]=='Z'?breakpoint_add:breakpoint_del)(x);
      echo(Bits("OK"));
      break;
    }
    default: {
      echo(Bits(""));
      break;
    }
  }
}

void GDB_RSP::cmdDetach(const std::string& str) {
  run();
  echo(Bits("OK"));
}

void GDB_RSP::cmdReset(const std::string& str) {
  Bits zero(register_write.pinout().pins(), false);
  register_address(Bits(PC));
  register_write(zero);
  run(Bits(static_cast<uint8_t>(0)));
  
  // no response ???
}

void GDB_RSP::cmdContinue(const std::string& str) {
  const char* ptr = str.c_str() + 1;
  
  uint64_t address;
  if (fromHex(ptr, address) == 0) {
    register_address(Bits(PC));
    register_write(Bits(address));
  }
  
  if (str[0] == 's')
    run(Bits(static_cast<uint8_t>(1)));
  else
    run(Bits(static_cast<uint8_t>(0)));
  
  cmdStatus(str);
}

void GDB_RSP::toHex(std::string& str, const Bits& bits) {
  std::string out;
  static const char hex[] = "0123456789abcdef";
  
  if (bits.size() % 8 != 0)
    throw ProtocolViolation(this, "Register is not byte-aligned");
  
  int len = bits.size() / 8;
  for (int i = len-1; i >= 0; --i) {
    uint8_t byte = bits.get_uint8(i);
    str.push_back(hex[byte >> 4]);
    str.push_back(hex[byte & 0xf]);
  }
}


static int hexDigit(char c) {
  if (c >= '0' && c <= '9')
    return c-'0';
  else if (c >= 'a' && c <= 'f')
    return c-'a'+10;
  else if (c >= 'A' && c <= 'F')
    return c-'A'+10;
  else
    return -1;
}

int GDB_RSP::fromHex(const char *&ptr, uint64_t& x) {
  int ok = -1;
  x = 0;
  for (int got; (got = hexDigit(*ptr)) >= 0; ++ptr) {
    x <<= 4;
    x |= got;
    ok = 0;
  }
  return ok;
}

int GDB_RSP::fromHex(const char *&ptr, Bits& bits) {
  if (bits.size() % 8 != 0)
    throw ProtocolViolation(this, "Register is not byte-aligned");
  
  int len = bits.size() / 8;
  for (int i = len-1; i >= 0; --i) {
    uint8_t b;
    int got;
    
    if ((got = hexDigit(*ptr)) < 0) return -1;
    ++ptr;
    b = got << 4;
    if ((got = hexDigit(*ptr)) < 0) return -1;
    ++ptr;
    b |= got;
    
    bits.set_uint8(i, b);
  }
  return 0;
}
