#ifndef GDB_RSP_H
#define GDB_RSP_H

#include "common/component.h"
#include <queue>

class GDB_RSP : public Component {
  public:
    GDB_RSP(Interconnect* ic, int width, int registers,
            int PC, int SP, int SIG);
    
    void recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end);
    SendQueue::const_iterator send(SendQueue::const_iterator i, SendQueue::const_iterator end);
    
  public:
    MasterPort& packet;
    SlavePort& echo;
    
    SlavePort& register_address; /* 8-bit */
    SlavePort& register_write;   /* width -> void */
    SlavePort& register_read;    /* 0-bit -> width */
    
    SlavePort& memory_address;   /* width -> void  */
    SlavePort& memory_write;     /* width -> void  */
    SlavePort& memory_read;      /* 0-bit -> width */
    
    /* The breakpoint type:
     *  00 = execution breakpoint
     *  01 = read watchpoint
     *  10 = write watchpoint
     *  11 = access watchpoint
     *
     * add/del: address of breakpoint
     */
    SlavePort& breakpoint_type;
    SlavePort& breakpoint_add; /* width -> void */
    SlavePort& breakpoint_del; /* width -> void */
    
    /* Run processor: 1 = single-step, 0 = continue */
    SlavePort& run; /* 8-bit -> void */
    
  protected:
    uint8_t registers, PC, SP, SIG;
    std::queue<int> queue; // -1 means status
    
    // Process a packet
    void runCommand(const std::string& str);
    
    // in:  ?
    // out: T<2H signal><2h PC#>:<4h PC>;<2h SP#>:<4h SP>;
    void cmdStatus(const std::string& str);
    
    // in: qSupported
    // out: PacketSize=8192
    void cmdQuery(const std::string& str);
    
    // in:  g
    // out: <4h*registers reg[i]>
    void cmdRegGetAll(const std::string& str);
    
    // in:  G<4h*registers reg[i]>
    // out: OK | E22
    void cmdRegSetAll(const std::string& str);
    
    // in:  m<h addr>,<h len>
    // out: E22 | E14 | <2h*bytes result>
    void cmdMemRead(const std::string& str);
    
    // in:  M<h addr>,<h len>:<2h*bytes payload>
    // out: OK | E22
    void cmdMemWrite(const std::string& str);
    
    // in:  p<2h reg#>
    // out: E22 | <4h reg[i]>
    void cmdRegGet(const std::string& str);
    
    // in:  P<2h reg#>=<4h reg>
    // out: OK | E22
    void cmdRegSet(const std::string& str);
    
    // in:  [zZ][1234],<h* addr>,<h* type>
    // out: OK | E28 | E22
    void cmdBreak(const std::string& str);
    
    // in:  [rR]
    // out: cmdStatus()
    void cmdReset(const std::string& str);
    
    // in:  [cs]<h addr>
    // out: cmdStatus()
    void cmdContinue(const std::string& str);
    
    // in:  [Dk]...
    // out: OK
    void cmdDetach(const std::string& str);
    
    void toHex(std::string& str, const Bits& bits);
    int fromHex(const char *&ptr, Bits& bits);  // fixed width
    int fromHex(const char *&ptr, uint64_t& x); // variable width
};

#endif
