#ifndef HEAD_OF_CHAIN_H
#define HEAD_OF_CHAIN_H

#include "common/component.h"
#include "pinouts/pinouts.h"
#include <map>

/* Use this component to schedule commands to subordinate components.
 * By sending them via a common root, consistent order is assured.
 */

class HeadOfChain : public Component {
  public:
    HeadOfChain(Interconnect* ic);
    
    void recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end);
    SendQueue::const_iterator send(SendQueue::const_iterator i, SendQueue::const_iterator end);
    
    struct Control {
      public:
        SlavePort& port();
        
      protected:
        Slaves::iterator it;
      
      friend class HeadOfChain;
    };
    
    /* Add a controller.
     * After connecting it to some component's master, you may send.
     * Any incoming data received will be dispatched to the provided callback.
     * If no callback is provided, throw a ProtocolViolation on recv.
     */
    template <typename T>
    Control addControl(const Pinout& po, T* user, void (*cb)(T*, const Bits&));
    Control addControl(const Pinout& po = command_pinout);
    void removeControl(Control c);
  
  protected:
    typedef void (*callback)(void*, const Bits&);
    struct CallbackInfo {
      callback cb;
      void*    user;
    };
    
    typedef std::map<SlavePort*, CallbackInfo> Map;
    
    unsigned int controlId;
    Map map;
    
    static void explode(HeadOfChain* obj, const Bits& b);
    Control addControlVoid(const Pinout& po, void* user, callback cb);
    SlavePort& addReset();
  
  friend class Interconnect;
};

/* Convert a member function like: 
 *  addControl(po, this, proxy<MyClass,&MyClass:member>);
 */
template <typename T, void (T::*cb)(const Bits&)>
void proxy(T* object, const Bits& bits) {
  return (object->*cb)(bits);
}

template <typename T>
inline HeadOfChain::Control HeadOfChain::addControl(const Pinout& po, T* user, void (*cb)(T*, const Bits&)) {
  return addControlVoid(po, user, reinterpret_cast<callback>(cb));
}

inline SlavePort& HeadOfChain::Control::port() {
  return *it;
}

#endif
