#include "common/compat.h"

#include "pinouts/pinouts.h"
#include "jtag-bitbang.h"
#include "common/to_string.h"

JTAG_BitBang::JTAG_BitBang(Interconnect* ic) 
 : Component(ic),
   IR_write       (*addMasterPort("IR/write",        shift_pinout)),
   IR_read_write  (*addMasterPort("IR/read_write",   shift_pinout)),
   DR_write       (*addMasterPort("DR/write",        shift_pinout)),
   DR_read_write  (*addMasterPort("DR/read_write",   shift_pinout)),
   JTAG_write     (*addSlavePort ("JTAG/write",      JTAG_pinout)),
   JTAG_read_write(*addSlavePort ("JTAG/read_write", JTAG_pinout)),
   TMS(JTAG_pinout.getPin("TMS")),
   TDI(JTAG_pinout.getPin("TDI")),
   TDO(JTAG_pinout.getPin("TDO")),
   queue() { }

void JTAG_BitBang::write(int tms, int tdi) {
  Bits b(JTAG_pinout.pins(), false);
  
  Bits::reference btdi = b[TDI];
  Bits::reference btms = b[TMS];
  
  btdi = tdi;
  btms = tms;
  
  JTAG_write(b);
}

void JTAG_BitBang::read_write(int tms, int tdi) {
  Bits b(JTAG_pinout.pins(), false);
  
  Bits::reference btdi = b[TDI];
  Bits::reference btms = b[TMS];
  
  btdi = tdi;
  btms = tms;
  
  JTAG_read_write(b);
}

SendQueue::const_iterator JTAG_BitBang::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  while (i != end) {
    if (i->master == &reset) {
      // 5 times send TMS=1 -> brings us to reset state
      for (int j = 0; j < 5; ++j) write(1, 0);
      // idle state
      write(0, 0);
      
      ++i;
      continue;
    }
    
    bool ir, read;
    if (i->master == &DR_read_write) {
      ir = false;
      read = true;
    } else if (i->master == &DR_write) {
      ir = false;
      read = false;
    } else if (i->master == &IR_read_write) {
      ir = true;
      read = true;
    } else {
      ir = true;
      read = false;
    }
    
    // Move to shift DR/IR state from idle state
    if (ir) {
      write(1, 0); // select DR scan
      write(1, 0); // select IR scan
      write(0, 0); // capture IR
      write(0, 0); // shift IR state
    } else { // else dr
      write(1, 0); // select DR scan
      write(0, 0); // capture DR
      write(0, 0); // shift DR state
    }
    
    // Shift in the bits
    Bits::const_iterator next, endb = i->bits.end();
    for (Bits::const_iterator j = i->bits.begin(); j != endb; j = next) {
      // Look-ahead increment
      next = j;
      ++next;
      
      // shift in new value
      if (read) {
        read_write(next == endb, *j);
      } else {
        write(next == endb, *j);
      }
    }
    
    if (read)
     queue.push(Op(ir, i->bits.size()));
      
    // now in exit-IR/DR state
    write(1, 0); // update IR/DR
    if (++i == end)
      write(0, 0); // run-test / idle 
      // ^^^ this last step CAN be optimized away (Altera does it for example)
      // However, some devices might hang expecting a completed update clock tick
      // Thus, only make the optimization in the middle of a command pipeline
  }
  
  return i;
}

void JTAG_BitBang::recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) {
  while (i != end) {
    if (queue.empty())
      throw ProtocolViolation(this, "More received data than expected");
    
    Op& op = queue.front();
    
    Bits result(op.bits, false);
    Bits::iterator endb = result.end();
    for (Bits::iterator j = result.begin(); j != endb; ++j) {
      if (i == end)
        throw ProtocolViolation(this, "Incomplete register transfer, expected: " + to_string(op.bits));
      *j = i->bits[TDO];
      ++i;
    }
    
    if (op.ir)
      IR_read_write(result);
    else
      DR_read_write(result);
    
    queue.pop();
  }
}
