#include "common/compat.h"

#include "jtag-chain.h"
#include "pinouts/pinouts.h"

#include <assert.h>

JTAG_Chain::JTAG_Chain(Interconnect* ic, int ml, int md)
 : Bus(ic), 
   IR_write     (*addSlavePort("IR/write",      shift_pinout)),
   IR_read_write(*addSlavePort("IR/read_write", shift_pinout)),
   DR_write     (*addSlavePort("DR/write",      shift_pinout)),
   DR_read_write(*addSlavePort("DR/read_write", shift_pinout)),
   max_irlen(ml),
   max_devices(md),
   ir_len(-1) {
}

void JTAG_Chain::recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) {
  while (i != end) {
    if (queue.empty())
      throw ProtocolViolation(this, "recv past queue end");
    
    MasterPort* master = queue.front();
    queue.pop();
    
    if (master == 0) {
      if (ir_len == -1)
        scanR(i);
      continue;
    }
    
    const std::string& name = master->name();
    uint32_t dev = device(name);
    bool ir = name[prefix] == 'I';
    assert(name.size() > (prefix+8));
    
    if (ir) {
      // Locate the IR inside the recv
      Bits result(devices[dev].ir_len, false);
      Bits::const_iterator j = i->bits.begin() + devices[dev].ir_offset;
      // Copy the bits
      Bits::iterator end = result.end();
      for (Bits::iterator o = result.begin(); o != end; ++o)
        *o = *j++;
      
      (*master)(result);
    } else {
      // Extract the data bits from the padding
      Bits result(i->bits.size() - devices.size()+1, false);
      // Fill starting past the predecessors
      Bits::const_iterator j = i->bits.begin() + dev;
      // Copy the bits
      Bits::iterator end = result.end();
      for (Bits::iterator o = result.begin(); o != end; ++o)
         *o = *j++;
      
      (*master)(result);
    }
    
    ++i;
  }
}

SendQueue::const_iterator JTAG_Chain::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  // Never blocks on read
  for (; i != end; ++i) {
    if (i->master == &reset) {
      if (ir_len == -1) {
        scanW();
        return i; // block
      }
      // already initialized
      continue; 
    }
    
    assert (!devices.empty()); // else could never have been called!
    if (devices[0].ir_len == -1)
      configWidths();
    
    const std::string& name = i->master->name();
    uint32_t dev = device(name);
    bool ir = name[prefix] == 'I';
    bool read = name.size() > (prefix+8);
    
    if (read) queue.push(i->master);
    
    if (ir) {
      // Build a bitvector with the appropriate IR for the device and BYPASS (1) for others.
      Bits command(ir_len, true);
      // Fill at the correct offset
      Bits::iterator o = command.begin() + devices[dev].ir_offset;
      // Copy the bits
      Bits::const_iterator bend = i->bits.end();
      for (Bits::const_iterator j = i->bits.begin(); j != bend; ++j)
        *o++ = *j;
      
      // Set it
      if (read)
        IR_read_write(command);
      else
        IR_write(command);
    } else {
      // Pad the DR bits with the BYPASS bits for inactive devices
      Bits padded(i->bits.size() + devices.size()-1, false);
      // Fill starting past the predecessors
      Bits::iterator o = padded.begin() + dev;
      // Copy the bits
      Bits::const_iterator bend = i->bits.end();
      for (Bits::const_iterator j = i->bits.begin(); j != bend; ++j)
        *o++ = *j;
      
      // Write out the padded command
      if (read)
        DR_read_write(padded);
      else
        DR_write(padded);
    }
  }
  
  return i;
}

JTAG_Chain::IDCODEs JTAG_Chain::detected() const {
  IDCODEs out(devices.size());
  
  for (Devices::size_type i = 0; i < devices.size(); ++i)
    out[i] = devices[i].code;
  return out;
}

JTAG_Chain::Device::Device(JTAG_Chain* owner, uint32_t dev) 
 : code(0),
   ir_len(-1),
   ir_offset(-1) {
  owner->addMasterPort(JTAG_Chain::makeName(dev, "IR/write"),      shift_pinout);
  owner->addMasterPort(JTAG_Chain::makeName(dev, "IR/read_write"), shift_pinout);
  owner->addMasterPort(JTAG_Chain::makeName(dev, "DR/write"),      shift_pinout);
  owner->addMasterPort(JTAG_Chain::makeName(dev, "DR/read_write"), shift_pinout);
}

void JTAG_Chain::scanW() {
  Bits idcode(max_devices*32, false);
  Bits ones  (max_irlen+1,    true);
  Bits zeros (max_devices+1,  false);
  
  // Set the magic flag to detect the run's end
  ones [0] = false;
  zeros[0] = true;
  
  // Grab the IDCODEs
  DR_read_write(idcode);
  
  // Put the IR into BYPASS mode (and record old IR = idcode_command)
  IR_read_write(ones);
  
  // Push zeros through to measure the number of devices
  DR_read_write(zeros);
  queue.push(0);
}

void JTAG_Chain::scanR(RecvQueue::const_iterator& i) {
  Bits::const_iterator j;
  
  assert (i->slave == &DR_read_write);
  const Bits& idcodes = i->bits;
  ++i;
  assert (i->slave == &IR_read_write);
  const Bits& irlenreg = i->bits;
  ++i;
  assert (i->slave == &DR_read_write);
  const Bits& devicesreg = i->bits;
  ++i;
  
  // Find the last occurance of 0 (aka: the first bit we shifted)
  for (j = irlenreg.end()-1; j != irlenreg.begin(); --j)
    if (*j == false) 
      break;
  
  ir_len = j - irlenreg.begin();
  if (ir_len <= 0)
    throw ProtocolViolation(this, "init failed to measure the length of the IR");
  
  // Find the last occurance of 1 (aka: the first bit we shifted)
  for (j = devicesreg.end()-1; j != devicesreg.begin(); --j)
    if (*j == true)
      break;
  
  int numdevices = j - devicesreg.begin();
  if (numdevices <= 0)
    throw ProtocolViolation(this, "init failed to measure the length of the IR");

  for (int k = 0; k < numdevices; ++k)
    devices.push_back(Device(this, k));
  
  // std::cout << "Saw " << numdevices << " devices with IR len " << ir_len << std::endl;
  for (unsigned int k = 0; k < devices.size(); ++k) {
    uint32_t code = 0;
    // Convert to unit32_t, high bits first
    for (int l = 31; l >= 0; --l) {
      code <<= 1;
      code |= idcodes[32*k+l]?1:0;
    }
    
    devices[k].code = code;
  }
}

void JTAG_Chain::configWidths() {
  int total = 0;
  for (Devices::size_type i = 0; i < devices.size(); ++i) {
    Masters::iterator m = findMaster(makeName(i, "IR/write"));
    if (!m->source())
      throw ProtocolViolation(this, "Device with IR/write disconnected");
    int len = m->source()->pinout().pins();
    
    devices[i].ir_offset = total;
    devices[i].ir_len = len;
    total += len;
  }
  
  if (total != ir_len)
    throw ProtocolViolation(this, "Device configuration does not fit IR width");
}
