#include "common/compat.h"

#include "lm32-drom.h"
#include "pinouts/pinouts.h"

#include <cstring>
#include <assert.h>

LM32DROM::LM32DROM(Interconnect* ic)
 : Component(ic),
   register_address(*addMasterPort("register/address", Pinout("address",  8))),
   register_write  (*addMasterPort("register/write",   Pinout("data",    32))),
   register_read   (*addMasterPort("register/read",    Pinout("data",    32))),
   memory_address  (*addMasterPort("memory/address",   Pinout("address", 32))),
   memory_write    (*addMasterPort("memory/write",     Pinout("data",     8))),
   memory_read     (*addMasterPort("memory/read",      Pinout("data",     8))),
   breakpoint_type (*addMasterPort("breadkpoint/type", Pinout("type",     8))),
   breakpoint_add  (*addMasterPort("breadkpoint/add",  Pinout("address", 32))),
   breakpoint_del  (*addMasterPort("breadkpoint/del",  Pinout("address", 32))),
   run             (*addMasterPort("run",              command_pinout)),
   echo      (*addSlavePort("echo",       uint32_pinout)),
   uart_write(*addSlavePort("uart/write", uint8_pinout)),
   uart_read (*addSlavePort("uart/read",  uint8_pinout)),
   uart_poll (*addSlavePort("uart/poll",  uint8_pinout)),
   jtag_break(*addSlavePort("uart/poll",  command_pinout, true)),
   state(RUNNING), needStop(true), needRead(false) {
}

void LM32DROM::stop() {
  needStop = true;
}

SendQueue::const_iterator LM32DROM::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (i->master == &reset) {
      state = RUNNING;
      needStop = true;
      needRead = false;
      registers = 0;
      memset(now, 0, sizeof(now));
      memset(old, 0, sizeof(now));
      continue;
    }
    
    switch (state) {
      case RUNNING: {
        if (needStop) {
          jtag_break();
          needStop = false;
        }
        if (needRead) {
          uart_read();
          needRead = false;
        }
        uart_poll();
        return i;
      }
      case LOAD_BASE: {
        uart_write(Bits(static_cast<uint8_t>(0x80)));
        uart_read();
        uart_read();
        uart_read();
        uart_read();
        return i;
      }
      case LOAD_REGS: {
        /* don't read R0 */
        read(registers+4, (RW_regs+R_regs-1)*4); 
        return i;
      }
      case STOPPED: {
        if (i->master == &register_address) {
          activeReg = static_cast<uint8_t>(i->bits);
        } else if (i->master == &register_read) {
          echo(Bits(getReg(activeReg)));
        } else if (i->master == &register_write) {
          setReg(activeReg, static_cast<uint32_t>(i->bits));
        } else if (i->master == &memory_address) {
          activeMem = i->bits;
        } else if (i->master == &memory_read) {
          uint32_t len;
          for (len = 0; i != end && i->master == &memory_read; ++i) 
            ++len;
          read(activeMem, len);
          --i;
        } else if (i->master == &memory_write) {
          write(activeMem, i, end);
          --i;
        } else if (i->master == &breakpoint_type) {
          activeType = static_cast<uint8_t>(i->bits);
        } else if (i->master == &breakpoint_add) {
          addBreak(static_cast<uint32_t>(i->bits));
        } else if (i->master == &breakpoint_del) {
          delBreak(static_cast<uint32_t>(i->bits));
        } else {
          assert (i->master == &run);
          if (i->bits[0])
            setReg(DC, getReg(DC) | 1);
          else
            setReg(DC, getReg(DC) & ~1);
          
          update();
          uart_write(Bits(static_cast<uint8_t>(0)));
          state = RUNNING;
        }
      }
    }
  }
  
  return i;
}

void LM32DROM::recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (i->slave == &echo) {
      register_read(i->bits);
    } else {
      switch (state) {
        case RUNNING: {
          if (i->slave == &uart_poll) {
            needRead = i->bits[0];
          } else if (i->slave == &uart_read) {
            uint8_t byte = i->bits;
            
            /* Peek at the poll result */
            if (++i == end || i->slave != &uart_poll)
              throw ProtocolViolation(this, "poll didn't follow read!");
            
            if (byte == 0x80 + 'B') {
              /* Should be last byte in buffer */
              if (i->bits[0] == false) {
                if (registers == 0)
                  state = LOAD_BASE;
                else
                  state = LOAD_REGS;
              } else {
              }
            } else {
              std::cerr << (char)byte << std::flush;
            }
          }
          break;
        }
        case LOAD_BASE: {
          registers = 
            static_cast<uint32_t>(static_cast<uint8_t>((i++)->bits)) << 24 |
            static_cast<uint32_t>(static_cast<uint8_t>((i++)->bits)) << 16 |
            static_cast<uint32_t>(static_cast<uint8_t>((i++)->bits)) <<  8 |
            static_cast<uint32_t>(static_cast<uint8_t>((i++)->bits)) <<  0;
          --i;
          state = LOAD_REGS;
          break;
        }
        case LOAD_REGS: {
          int pulled = (RW_regs + R_regs)*4;
          old[0] = old[1] = old[2] = old[3] = 0; /* R0 */
          for (int j = 4; j < pulled; ++i, ++j) /* Skip R0 */ {
            assert (i != end);
            assert (i->slave == &uart_read);
            old[j] = static_cast<uint8_t>(i->bits);
          }
          --i;
          memcpy(now, old, pulled);
          
          /* Fixup SIG and PC */
          switch ((getReg(SIG) & 0xff) >> 5) {
          case 0: /* reset */
            setReg(SIG, 0);
            setReg(PC, getReg(EA));
            break;
          case 1: /* breakpoint */
            setReg(SIG, 5); /* SIGTRAP */
            setReg(PC, getReg(BA));
            break;
          case 2: /* instruction bus error */
            setReg(SIG, 11); /* SIGSEGV */
            setReg(PC, getReg(EA));
            break;
          case 3: /* watchpoint */
            setReg(SIG, 5); /* SIGTRAP */
            setReg(PC, getReg(BA));
            break;
          case 4: /* data bus error */
            setReg(SIG, 11); /* SIGSEGV */
            setReg(PC, getReg(EA));
            break;
          case 5: /* divide by zero */
            setReg(SIG, 8); /* SIGFPE */
            setReg(PC, getReg(EA));
            break;
          case 6: /* interrupt */
            setReg(SIG, 2); /* SIGINT */
            setReg(PC, getReg(EA));
            break;
          case 7: /* system call */
            setReg(SIG, 1); /* SIGHUP */
            setReg(PC, getReg(EA));
            break;
          }
          
          state = STOPPED;
          break;
        }
        case STOPPED: {
          memory_read(i->bits);
          break;
        }
      }
    }
  }
}

void LM32DROM::read(uint32_t address, uint32_t len) {
  while (len > 0) {
    uint32_t cut;
    if (len > 127) cut = 127; else cut = len;
    
    uart_write(Bits(static_cast<uint8_t>(0x80 | cut)));
    uart_write(Bits(static_cast<uint8_t>(address >> 24)));
    uart_write(Bits(static_cast<uint8_t>(address >> 16)));
    uart_write(Bits(static_cast<uint8_t>(address >>  8)));
    uart_write(Bits(static_cast<uint8_t>(address >>  0)));
    for (uint32_t j = 0; j < cut; ++j)
      uart_read();
    
    address += cut;
    len -= cut;
  }
}

void LM32DROM::write(uint32_t address, SendQueue::const_iterator& i, SendQueue::const_iterator end) {
  uint32_t len = 0;
  for (SendQueue::const_iterator scan = i; scan != end && scan->master == &memory_write; ++scan) 
    ++len;
  
  while (len > 0) {
    uint32_t cut;
    if (len > 127) cut = 127; else cut = len;
    
    uart_write(Bits(static_cast<uint8_t>(cut)));
    uart_write(Bits(static_cast<uint8_t>(address >> 24)));
    uart_write(Bits(static_cast<uint8_t>(address >> 16)));
    uart_write(Bits(static_cast<uint8_t>(address >>  8)));
    uart_write(Bits(static_cast<uint8_t>(address >>  0)));
    for (uint32_t j = 0; j < cut; ++j, ++i)
      uart_write(i->bits);
    
    address += cut;
    len -= cut;
  }
}

void LM32DROM::setReg(int reg, uint32_t val) {          
  now[reg*4+0] = val >> 24;
  now[reg*4+1] = val >> 16;
  now[reg*4+2] = val >>  8;
  now[reg*4+3] = val >>  0;
}

uint32_t LM32DROM::getReg(int reg) {
  return
    static_cast<uint32_t>(now[reg*4+0]) << 24 |
    static_cast<uint32_t>(now[reg*4+1]) << 16 |
    static_cast<uint32_t>(now[reg*4+2]) <<  8 |
    static_cast<uint32_t>(now[reg*4+3]) <<  0;
}

void LM32DROM::addBreak(uint32_t address) {
  uint32_t cfg = getReg(CFG);
  int bpoints = (cfg >> 18) & 0xf;
  int wpoints = (cfg >> 22) & 0xf;
  
  if (activeType == 0) {
    /* Find a breakpoint slot */
    for (int i = 0; i < bpoints; ++i) {
      if ((getReg(BP0 + i) & 1) == 0) {
        setReg(BP0 + i, address|1);
        break;
      }
    }
  } else {
    /* Find a watchpoint slot */
    int mask = 0xc;
    int type = activeType << 2;
    for (int i = 0; i < wpoints; ++i) {
      if ((getReg(DC) & mask) == 0) {
        setReg(DC, getReg(DC) | type);
        setReg(WP0 + i, address);
        break;
      }
      mask <<= 2;
      type <<= 2;
    }
  }
}

void LM32DROM::delBreak(uint32_t address) {
  uint32_t cfg = getReg(CFG);
  int bpoints = (cfg >> 18) & 0xf;
  int wpoints = (cfg >> 22) & 0xf;
  
  if (activeType == 0) {
    /* Find a breakpoint slot */
    for (int i = 0; i < bpoints; ++i) {
      if (getReg(BP0 + i) == (address|1)) {
        setReg(BP0 + i, 0);
        break;
      }
    }
  } else {
    /* Find a watchpoint slot */
    int mask = 0xc;
    int type = activeType << 2;
    for (int i = 0; i < wpoints; ++i) {
      if (getReg(WP0 + i) == address) {
        setReg(DC, getReg(DC) & ~mask);
        setReg(WP0 + i, 0);
        break;
      }
      mask <<= 2;
      type <<= 2;
    }
  }
}

void LM32DROM::update() {
  now[0] = now[1] = now[2] = now[3] = 0; /* R0 */
  
  /* Copy PC into place */
  switch (getReg(SIG)) {
  case 5:  setReg(BA, getReg(PC)); break;
  default: setReg(EA, getReg(PC)); break;
  }
  
#if 0
  /* Analyze the longest run-length needing update */
  int runLen[regs*4];
  for (int j = regs*4-1; j >= 0; --j
#endif

  /* !!! For now, write it all */
    uart_write(Bits(static_cast<uint8_t>(100)));
    uart_write(Bits(static_cast<uint8_t>(registers >> 24)));
    uart_write(Bits(static_cast<uint8_t>(registers >> 16)));
    uart_write(Bits(static_cast<uint8_t>(registers >>  8)));
    uart_write(Bits(static_cast<uint8_t>(registers >>  0)));
    for (uint32_t j = 0; j < 100; ++j)
      uart_write(Bits(now[j]));
    
    uart_write(Bits(static_cast<uint8_t>(100)));
    uart_write(Bits(static_cast<uint8_t>((registers+100) >> 24)));
    uart_write(Bits(static_cast<uint8_t>((registers+100) >> 16)));
    uart_write(Bits(static_cast<uint8_t>((registers+100) >>  8)));
    uart_write(Bits(static_cast<uint8_t>((registers+100) >>  0)));
    for (uint32_t j = 100; j < 200; ++j)
      uart_write(Bits(now[j]));
}
