#include "common/compat.h"

// generic components
#include "common/interconnect.h"
#include "components/echo/echo.h"
#include "components/arbiter/arbiter.h"
#include "components/map/map.h"
#include "components/reduce/reduce.h"
// needed for lm32 on PCIe
#include "components/rawrabbit/rawrabbit.h"
// LM32 contols
#include "components/lm32/lm32-break.h"
#include "components/lm32/lm32-reboot.h"
#include "components/lm32/lm32-read.h"
#include "components/lm32/lm32-write.h"
#include "components/lm32/lm32-drom.h"
// GDB 
#include "components/gdb-rsp/gdb-rsp.h"
#include "components/gdb-rsp/gdb-serial.h"

#include <iostream>
#include <iomanip>
#include <usb.h>
#include <stdio.h>
#include <signal.h>

void got(void* x, const Bits& b) {
  uint8_t byte = static_cast<uint8_t>(b);
  std::cout << (char)byte << std::flush;
}

LM32DROM* drom = 0;
void trapSIGINT(int sig) {
  drom->stop();
}

int main() {
  Interconnect ic;
  
  Echo echo(&ic);
  Arbiter echos(&ic, true);
  RawRabbit rabbit(&ic);
  Arbiter dev0_a(&ic, false);
  Map dev0_w(&ic, 32, false);
  Map dev0_r(&ic, 32, true);
  
  Arbiter lm32_w(&ic, false);
  Arbiter lm32_r(&ic, true);
  LM32Break lm32break(&ic);
  LM32Reboot lm32reboot(&ic);
  LM32Write lm32write(&ic, true);
  LM32Read lm32read(&ic, true);
  LM32DROM lm32drom(&ic);
  GDB_RSP gdb_rsp(&ic, 32, 32, lm32drom.PC, lm32drom.SP, lm32drom.SIG);
  GDB_Serial gdb_serial(&ic);
  
  SlavePort& uart = ic.head().addControl(uint8_pinout, (void*)0, &got).port();
  
  try {
    ic.configure()
      .connect(echo.effect, ic.end().addEffect().port())
      .connect(echos.slave, echo.echo)
      .connect(rabbit.effect, ic.end().addEffect().port())
      .connect(dev0_a.slave, rabbit.address)
      .connect(dev0_w.address, dev0_a.addClient().port())
      .connect(dev0_w.data, rabbit.write)
      .connect(dev0_r.address, dev0_a.addClient().port())
      .connect(dev0_r.data, rabbit.read)
    
      .connect(lm32_w.slave, dev0_w.addRegister(0x70000).port())
      .connect(lm32_r.slave, dev0_r.addRegister(0x70000).port())
    
      .connect(lm32break.lm32_write, lm32_w.addClient().port())
      .connect(lm32break.lm32_read, lm32_r.addClient().port())
      .connect(lm32reboot.lm32_write, lm32_w.addClient().port())
      .connect(lm32reboot.lm32_read, lm32_r.addClient().port())
      .connect(lm32write.lm32_write, lm32_w.addClient().port())
      .connect(lm32write.lm32_read, lm32_r.addClient().port())
      .connect(lm32read.lm32_write, lm32_w.addClient().port())
      .connect(lm32read.lm32_read, lm32_r.addClient().port())
      .connect(lm32drom.echo, echos.addClient().port())
      .connect(lm32drom.uart_write, lm32write.write)
      .connect(lm32drom.uart_read, lm32read.read)
      .connect(lm32drom.uart_poll, lm32read.poll)
      .connect(lm32drom.jtag_break, lm32break.execute)
      
      .connect(gdb_rsp.echo, echos.addClient().port())
      .connect(gdb_rsp.register_address, lm32drom.register_address)
      .connect(gdb_rsp.register_write, lm32drom.register_write)
      .connect(gdb_rsp.register_read, lm32drom.register_read)
      .connect(gdb_rsp.memory_address, lm32drom.memory_address)
      .connect(gdb_rsp.memory_write, lm32drom.memory_write)
      .connect(gdb_rsp.memory_read, lm32drom.memory_read)
      .connect(gdb_rsp.breakpoint_type, lm32drom.breakpoint_type)
      .connect(gdb_rsp.breakpoint_add, lm32drom.breakpoint_add)
      .connect(gdb_rsp.breakpoint_del, lm32drom.breakpoint_del)
      .connect(gdb_rsp.run, lm32drom.run)
      .connect(gdb_serial.echo, echos.addClient().port())
      .connect(gdb_serial.packet, gdb_rsp.packet)
      .connect(uart, gdb_serial.uart)
      ;
  } catch (const BadConnection& bc) {
    std::cerr << bc;
    return -1;
  }
  
  try {
    ic.reset();
    ic.execute();
  } catch (const ProtocolViolation& v) {
    std::cerr << v;
    return -1;
  }
  
  drom = &lm32drom;
  signal(SIGINT, &trapSIGINT);
  
  std::cerr << "Executing break -- and entering pipe-mode" << std::endl;
  
  uint8_t buf[1024];
  while (1) {
    // Poll device
    try {
      ic.execute();
    } catch (const ProtocolViolation& v) {
      std::cerr << v;
      return -1;
    }
    
    // Get next command
    int got = read(0, buf, sizeof(buf));
    if (got <= 0) break; // EOF
    for (int i = 0; i < got; ++i)
      uart(Bits(buf[i]));
  }
  
  return 0;
}
